require 'rspec/core'
require 'yaml'

class Object
  def self.define_truth &block
    DDT::TruthTest::define_truth_for :class, self, &block
  end
end

class Module
  def define_truth &block
    DDT::TruthTest::define_truth_for :module, self, &block
  end
end

# @private
module DDT
  module TruthTest

    def define_truth_for obj_type, the_klass, &block
      klass = Class.new

      klass.class_eval do
        # defined so we can override it with should_receive
        def self.tester
          nil
        end

        attr_accessor :input, :output
      end

      if block_given?
        pblock = block
      else
        pblock = nil
      end

      # this is a bit of trickery to get at the private method
      #  :define_method and also allow access to the_klass
      klass.send(:define_method, :initialize) do |data={}|
        data.each_pair do |attr, value|
          next if attr == '!raises'
          instance_variable_set "@#{attr}", value
          self.class.send(:attr_accessor, attr.to_sym)
        end

        instance_variable_set "@input", data["input"]
        (instance_variable_set "@output", data["output"]) if data["output"]

        if pblock
          pblock.call(self, data)
        end
      end

      klass.send(:define_method, :tester) do
        if obj_type == :class
          self.class.send(:tester) || the_klass.new 
        elsif obj_type == :module
          self.class.send(:tester) || the_klass
        end
      end

      the_klass.const_set "Truth", klass

    end

    module_function :define_truth_for

    class RSpec::Core::ExampleGroup

      # Declare a truth test for a method of the class being described.
      # If the class did not have define_truth called, it will be
      # called automatically with no arguments.
      #
      # Example
      #     # truths/foo/do_stuff_truths.yaml
      #     ---
      #     input: blah
      #     bar: baz blah
      #     ---
      #     input: bam
      #     bar: baz bam
      #     ---
      #     input: uh-oh
      #     '!raises': true
      #
      #     # foo.rb
      #     class Foo
      #       attr_accessor :bar
      #
      #       def do_stuff input
      #         if input == 'uh-oh'
      #           raise ArgumentError, 'This input raises an error'
      #         end
      #         @bar = "baz #{input}"
      #       end
      #     end
      #
      #     # foo_spec.rb
      #     Foo.define_truth
      #     
      #     describe Foo do
      #       truth_test :do_stuff
      #     end
      #
      # opts can either be an options hash or a filename in which to find the
      # truth data.  If opts is a hash and it includes :data as a key, the 
      # value of opts :data is a string to be parsed as YAML.  If opts 
      # includes the :truth_file, then that file is used for truth (same
      # as using a string instead of a hash).
      #
      # If no input is specified (i.e., if opts is omitted), then truth_test will
      # look for a yaml file called ./truths/class_name/method_truths.yaml 
      # (class_name in snake case a la Rails).
      def self.truth_test method, opts={}
        unless described_class.const_defined?("Truth")
          described_class.define_truth
        end

        if (opts.is_a? Hash) && opts[:data]
          data = opts[:data]
        else
          filename = opts if opts.is_a? String
          filename ||= (opts[:truth_file] || "./truths/#{DDT::underscore(described_class.name)}/#{method}_truths.yaml")
          data = File.open(filename, "rb")
        end

        run_truth_test method, data
      ensure
        data.close if data.is_a? File
      end

      # @private
      def self.run_truth_test method, data
        truth_count = -1
        YAML::load_documents(data).each do |datum|
          truth_count += 1
          truth = described_class::Truth.new datum

          # don't use the value because we might want nil input
          unless datum.has_key?("input")
            puts "Skipping truth datum #{truth_count} with no input..."
            next
          end

          # test for an exception to be raised if the datum has a '!raise'
          #   key that is set to true
          if datum.has_key?('!raises')
            if datum['!raises'] == true 
              # true -> raise any kind of error
              it "should raise an error when called as '#{method}(#{truth.input})'" do
                expect{ truth.tester.send(method, truth.input) }.to raise_error
              end
              next
            else 
              # specific value -> interpret that as an error class
              it "should raise an error of type #{datum['!raises']} when called as '#{method}(#{truth.input})'" do
                expect{ truth.tester.send(method, truth.input) }.to raise_error(Kernel.const_get(datum['!raises']))
              end
              next
            end
          end

          if datum.has_key?("output")
            it "should respond to '#{method}(#{truth.input})' with '#{truth.output}'" do
              truth.tester.send(method, truth.input).should == truth.output
            end
          end

          # already tested these
          datum.delete("input")
          datum.delete("output")

          datum.each_key do |attr|
            it "when the input is '#{truth.input}', calling '#{attr}' should return '#{truth.send(attr)}'" do
              obj = truth.tester
              obj.send(method, truth.input)
              obj.send(attr).should == truth.send(attr)
            end
          end
        end
      end

    end
  end
end
